/***************************************************************************
 *                                                                         *
 *      Copyright (c) 2009 by Dirk Clemens <develop@cle-mens.de>           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   See file gpl-2.0.txt or http://www.gnu.org/licenses/gpl-2.0.txt       *
 *                                                                         *
 ***************************************************************************/

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>

#include "debug.h"
#include "version.h"
#include "types.h"
#include "lib-wdf.h"
#include "lib-std.h"

///////////////////////////////////////////////////////////////////////////////

#define NAME "wdf-cat"
#define TITLE NAME " v" VERSION " r" REVISION " " SYSTEM " - " AUTHOR " - " DATE

char zerobuf[0x100000];

//
///////////////////////////////////////////////////////////////////////////////

enum // const for long options without a short brothers
{
	GETOPT_BASE	= 0x1000,
	GETOPT_IO,
};

char short_opt[] = "hV";
struct option long_opt[] =
{
	{ "help",	0, 0, 'h' },
	{ "version",	0, 0, 'V' },

	{ "io",		1, 0, GETOPT_IO }, // [2do] hidden option for tests

	{0,0,0,0}
};

///////////////////////////////////////////////////////////////////////////////

static char help_text[] =
    "\n"
    TITLE "\n"
    "This tool concatenate files and print the result to standard output.\n"
    "WDF files will be expanded to their normal presentation.\n"
    "\n"
    "Syntax: " NAME " [option]... files...\n"
    "\n"
    "Options:\n"
    "\n"
    "    -h --help     Print this help and exit.\n"
    "    -V --version  Print program name and version and exit.\n"
    "\n";

///////////////////////////////////////////////////////////////////////////////

void help_exit()
{
    fputs(help_text,stdout);
    exit(ERR_OK);
}

///////////////////////////////////////////////////////////////////////////////

void version_exit()
{
    printf("%s\n",TITLE);
    exit(ERR_OK);
}

///////////////////////////////////////////////////////////////////////////////

void hint_exit ( int stat )
{
    fprintf(stderr,
	    "-> Type '%s -h' or %s --help for more help.\n\n",
	    progname, progname );
    exit(stat);
}

///////////////////////////////////////////////////////////////////////////////

enumError CheckOptions ( int argc, char ** argv )
{
    int err = 0;
    for(;;)
    {
	const int opt_stat = getopt_long(argc,argv,short_opt,long_opt,0);
	if ( opt_stat == -1 )
	    break;

	noTRACE("CHECK OPTION %02x\n",opt_stat);
	switch (opt_stat)
	{
	  case '?': err++; break;
	  case 'V': version_exit();
	  case 'h': help_exit();

	  case GETOPT_IO:
	    {
		const enumIOMode new_io = strtol(optarg,0,0); // [2do] error handling
		opt_iomode = new_io & IOM__IS_MASK;
		if ( opt_iomode != new_io )
		    fprintf(stderr,"IO mode set to %#0x.\n",opt_iomode);
		opt_iomode |= IOM_FORCE_STREAM;
	    }
	    break;

	  default:
	    ERROR0(ERR_INTERNAL,"Internal error: unhandled option: '%c'\n",opt_stat);
	    ASSERT(0); // line never reached
	    break;
	}
    }
    return err ? ERR_SYNTAX : ERR_OK;
}

//
///////////////////////////////////////////////////////////////////////////////

enumError do_cat ( File_t * f, off_t src_off, u64 size64 )
{
    while ( size64 > 0 )
    {
	if ( SIGINT_level > 1 )
	    return ERR_INTERRUPT;

	u32 size = sizeof(iobuf);
	if ( size > size64 )
	    size = (u32)size64;

	int stat = ReadAtF(f,src_off,iobuf,size);
	if (stat)
	    return stat;

	size_t wstat = fwrite(iobuf,1,size,stdout);
	if ( wstat != size )
	    return ERR_WRITE_FAILED;

	src_off += size;
	size64  -= size;
    }
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

enumError do_zero ( u64 size64 )
{
    while ( size64 > 0 )
    {
	if ( SIGINT_level > 1 )
	    return ERR_INTERRUPT;

	u32 size = sizeof(iobuf);
	if ( size > size64 )
	    size = (u32)size64;

	size_t wstat = fwrite(zerobuf,1,size,stdout);
	if ( wstat != size )
	    return ERR_WRITE_FAILED;

	size64 -= size;
    }
    return ERR_OK;
}

///////////////////////////////////////////////////////////////////////////////

enumError wdf_cat ( ccp fname )
{
    SuperFile_t sf;
    InitializeSF(&sf);
    enumError stat = OpenFile(&sf.f,fname,IOM_IS_IMAGE);
    if (stat)
	return stat;

    stat = SetupReadWDF(&sf);
    if ( stat == ERR_NO_WDF )
    {
	// normal cat

	TRACE("raw copy %llu bytes.\n",sf.f.st.st_size);
	stat = do_cat(&sf.f,0,sf.f.st.st_size);
	CloseSF(&sf,0);
	return stat;
    }

    if (stat)
    {
	CloseSF(&sf,0);
	return stat;
    }

    int i;
    u64 last_off = 0;
    for ( i=0; i < sf.wc_used; i++ )
    {
	if ( SIGINT_level > 1 )
	{
	    stat = ERR_INTERRUPT;
	    break;
	}

	WDF_Chunk_t *wc = sf.wc + i;
	long long zero_count = wc->file_pos - last_off;
	if ( zero_count > 0 )
	{
	    TRACE("zero #%02d @%9llx [%9llx].\n",i,last_off,zero_count);
	    last_off += zero_count;
	    stat = do_zero(zero_count);
	    if (stat)
		return stat;
	}

	if ( wc->data_size )
	{
	    TRACE("copy #%02d @%9llx [%9llx] read-off=%9llx.\n",
		    i, last_off, wc->data_size, wc->data_off );
	    last_off += wc->data_size;
	    stat = do_cat(&sf.f,wc->data_off,wc->data_size);
	    if (stat)
		return stat;
	}
    }

    CloseSF(&sf,0);
    return stat;
}

///////////////////////////////////////////////////////////////////////////////

int main ( int argc, char ** argv )
{
    SetupLib(argc,argv,NAME,PROG_WDF_CAT);

    //----- process arguments

    if ( argc < 2 )
    {
	printf("\n%s\n\n",TITLE);
	hint_exit(ERR_OK);
    }

    if (CheckOptions(argc,argv))
	hint_exit(ERR_SYNTAX);

    memset(zerobuf,0,sizeof(zerobuf));

    int i, max_err = ERR_OK;
    for ( i = optind; i < argc && !SIGINT_level; i++ )
    {
	const int last_err = wdf_cat(argv[i]);
	if ( max_err < last_err )
	    max_err = last_err;
    }

    if (SIGINT_level)
	ERROR0(ERR_INTERRUPT,"Program interrupted by user.");
    return max_err > max_error ? max_err : max_error;
}

///////////////////////////////////////////////////////////////////////////////
